@@ -1,39 +1,13 @@
-# -*- encoding: utf-8 -*-
-
-#**********************************************************************************************************************#
-#                                        PYTOOLBOX - TOOLBOX FOR PYTHON SCRIPTS
-#
-#  Main Developer : David Fischer (david.fischer.ch@gmail.com)
-#  Copyright      : Copyright (c) 2012-2014 David Fischer. All rights reserved.
-#
-#**********************************************************************************************************************#
-#
-# This file is part of David Fischer's pytoolbox Project.
-#
-# This project is free software: you can redistribute it and/or modify it under the terms of the EUPL v. 1.1 as provided
-# by the European Commission. This project is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
-# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-#
-# See the European Union Public License for more details.
-#
-# You should have received a copy of the EUPL General Public License along with this project.
-# If not, see he EUPL licence v1.1 is available in 22 languages:
-#     22-07-2013, <https://joinup.ec.europa.eu/software/page/eupl/licence-eupl>
-#
-# Retrieved from https://github.com/davidfischer-ch/pytoolbox.git
-
-from __future__ import absolute_import, division, print_function, unicode_literals
-
 import struct
-from fastxor import fast_xor_inplace
 
-from ..rtp import RtpPacket
-from ...encoding import to_bytes
+from fastxor import fast_xor_inplace  # pylint:disable=no-name-in-module
 
-__all__ = ('FecPacket', )
+from pytoolbox.network.rtp import RtpPacket
+
+__all__ = ['FecPacket']
 
 
-class FecPacket(object):
+class FecPacket(object):  # pylint:disable=too-many-instance-attributes
     """
     This represent a real-time transport protocol (RTP) packet.
 
@@ -58,8 +32,8 @@ class FecPacket(object):
         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 
     The constructor will parse input bytes array to fill packet's fields.
-    In case of error (e.g. bad version number) the constructor will abort filling fields and un-updated fields are set
-    to their corresponding default value.
+    In case of error (e.g. bad version number) the constructor will abort filling fields and
+    un-updated fields are set to their corresponding default value.
 
     :param bytes: Input array of bytes to parse as a RTP packet with FEC payload
     :type bytes: bytearray
@@ -82,7 +56,7 @@ class FecPacket(object):
     >>> print(len(bytes))
     1344
     >>> fec = FecPacket(bytes, len(bytes))
-    >>> assert(fec.valid)
+    >>> assert fec.valid
     >>> print(fec)
     errors                = []
     sequence              = 37798
@@ -111,7 +85,7 @@ class FecPacket(object):
     >>> print(len(bytes))
     1344
     >>> fec = FecPacket(bytes, len(bytes))
-    >>> assert(fec.valid)
+    >>> assert fec.valid
     >>> print(fec)
     errors                = []
     sequence              = 63004
@@ -128,7 +102,7 @@ class FecPacket(object):
     missing               = []
     """
 
-    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constants >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constants >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
 
     ER_PAYLOAD_TYPE = 'RTP Header : Payload type must be set to 96'
     ER_EXTENDED = 'SMPTE 2022-1 Header : Extended must be set to one'
@@ -160,14 +134,14 @@ class FecPacket(object):
     SNBE_SHIFT = 16
 
     DIRECTION_NAMES = ('COL', 'ROW')
-    DIRECTION_RANGE = xrange(len(DIRECTION_NAMES))
+    DIRECTION_RANGE = range(len(DIRECTION_NAMES))  # noqa
     COL, ROW = DIRECTION_RANGE
 
     ALGORITHM_NAMES = ('XOR', 'Hamming', 'ReedSolomon')
-    ALGORITHM_RANGE = xrange(len(ALGORITHM_NAMES))
+    ALGORITHM_RANGE = range(len(ALGORITHM_NAMES))  # noqa
     XOR, Hamming, ReedSolomon = ALGORITHM_RANGE
 
-    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Properties >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Properties >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
 
     @property
     def valid(self):
@@ -189,31 +163,31 @@ class FecPacket(object):
         #TODO >>> print(fec.errors)
         #TODO ['RTP Header : Version must be set to 2', 'RTP packet must have a payload']
         """
-        errors = self._errors
+        errors = self._errors[:]
         if not self.extended:
-            errors.append(FecPacket.ER_EXTENDED)
+            errors.append(self.ER_EXTENDED)
         if self.mask != 0:
-            errors.append(FecPacket.ER_MASK)
+            errors.append(self.ER_MASK)
         if self.n:
-            errors.append(FecPacket.ER_N)
-        if self.algorithm != FecPacket.XOR:
-            errors.append(FecPacket.ER_ALGORITHM)
-        if self.direction not in FecPacket.DIRECTION_RANGE:
-            errors.append(FecPacket.ER_DIRECTION)
+            errors.append(self.ER_N)
+        if self.algorithm != self.XOR:
+            errors.append(self.ER_ALGORITHM)
+        if self.direction not in self.DIRECTION_RANGE:
+            errors.append(self.ER_DIRECTION)
         if self.index != 0:
-            errors.append(FecPacket.ER_INDEX)
+            errors.append(self.ER_INDEX)
         if self.payload_size == 0:
-            errors.append(FecPacket.ER_PAYLOAD)
+            errors.append(self.ER_PAYLOAD)
         if self.L < 1 or self.L > 50:
-            errors.append(FecPacket.ER_L)
-        if self.direction == FecPacket.COL and self.L * self.D > 256:
-            errors.append(FecPacket.ER_LD)
-        if self.direction == FecPacket.COL and (self.D < 4 or self.D > 50):
-            errors.append(FecPacket.ER_D)
+            errors.append(self.ER_L)
+        if self.direction == self.COL and self.L * self.D > 256:
+            errors.append(self.ER_LD)
+        if self.direction == self.COL and (self.D < 4 or self.D > 50):
+            errors.append(self.ER_D)
         return errors
 
     @property
-    def D(self):
+    def D(self):  # pylint:disable=invalid-name
         """
         Returns the vertical size of the FEC matrix (rows).
 
@@ -225,23 +199,23 @@ class FecPacket(object):
         >>> print(fec.D)
         2
         """
-        return self.na if self.direction == FecPacket.COL else None
+        return self.na if self.direction == self.COL else None
 
     @property
-    def L(self):
+    def L(self):  # pylint:disable=invalid-name
         """
         Returns the horizontal size of the FEC matrix (columns).
 
         **Example usage**
 
-        >>> from ..rtp import RtpPacket
+        >>> from pytoolbox.network.rtp import RtpPacket
         >>> packets = [RtpPacket.create(10, 100, RtpPacket.MP2T_PT, bytearray(123)),
         ...            RtpPacket.create(11, 200, RtpPacket.MP2T_PT, bytearray(1234))]
         >>> fec = FecPacket.compute(6, FecPacket.XOR, FecPacket.ROW, 2, 1, packets)
         >>> print(fec.L)
         2
         """
-        return self.offset if self.direction == FecPacket.COL else self.na
+        return self.offset if self.direction == self.COL else self.na
 
     @property
     def header_size(self):
@@ -250,14 +224,14 @@ class FecPacket(object):
 
         **Example usage**
 
-        >>> from ..rtp import RtpPacket
+        >>> from pytoolbox.network.rtp import RtpPacket
         >>> packets = [RtpPacket.create(10, 100, RtpPacket.MP2T_PT, bytearray(123)),
         ...            RtpPacket.create(11, 200, RtpPacket.MP2T_PT, bytearray(1234))]
         >>> fec = FecPacket.compute(1985, FecPacket.XOR, FecPacket.ROW, 2, 1, packets)
         >>> print(fec.header_size)
         16
         """
-        return FecPacket.HEADER_LENGTH
+        return self.HEADER_LENGTH
 
     @property
     def payload_size(self):
@@ -266,7 +240,7 @@ class FecPacket(object):
 
         **Example usage**
 
-        >>> from ..rtp import RtpPacket
+        >>> from pytoolbox.network.rtp import RtpPacket
         >>> packets = [RtpPacket.create(10, 100, RtpPacket.MP2T_PT, bytearray(123)),
         ...            RtpPacket.create(11, 200, RtpPacket.MP2T_PT, bytearray(1234))]
         >>> fec = FecPacket.compute(27, FecPacket.XOR, FecPacket.ROW, 2, 1, packets)
@@ -282,9 +256,11 @@ class FecPacket(object):
 
         **Example usage**
 
-        >>> from ..rtp import RtpPacket
-        >>> packets = [RtpPacket.create(10, 100, RtpPacket.MP2T_PT, bytearray(123)),
-        ...            RtpPacket.create(11, 200, RtpPacket.MP2T_PT, bytearray(1234))]
+        >>> from pytoolbox.network.rtp import RtpPacket
+        >>> packets = [
+        ...     RtpPacket.create(10, 100, RtpPacket.MP2T_PT, bytearray(123)),
+        ...     RtpPacket.create(11, 200, RtpPacket.MP2T_PT, bytearray(1234))
+        ... ]
         >>> fec = FecPacket.compute(26, FecPacket.XOR, FecPacket.ROW, 2, 1, packets)
         >>> print(fec)
         errors                = []
@@ -301,103 +277,107 @@ class FecPacket(object):
         payload recovery size = 1234
         missing               = []
         >>> fec_header = fec.header_bytes
-        >>> assert(len(fec_header) == FecPacket.HEADER_LENGTH)
-        >>> print(''.join(' %02x' % b for b in fec_header))
-         00 0a 04 a9 80 00 00 00 00 00 00 ac 40 01 02 00
+        >>> len(fec_header)  # FecPacket.HEADER_LENGTH
+        16
+        >>> ''.join(' %02x' % b for b in fec_header)
+        ' 00 0a 04 a9 80 00 00 00 00 00 00 ac 40 01 02 00'
         >>> fec_header += fec.payload_recovery
         >>> rtp = RtpPacket.create(26, 100, RtpPacket.DYNAMIC_PT, fec_header)
-        >>> rtp_header = rtp.header_bytes
-        >>> header = rtp_header + fec_header
-        >>> fec2 = FecPacket(header, len(header))
-        >>> fec_header = fec2.header_bytes
-        >>> assert(fec == fec2)
+        >>> header = rtp.header_bytes + fec_header
+        >>> fec == FecPacket(header, len(header))
+        True
         """
         # FIXME map type string to enum
-        bytes = bytearray(FecPacket.HEADER_LENGTH)
-        struct.pack_into(b'!H', bytes, 0, self.snbase & FecPacket.SNBL_MASK)
-        struct.pack_into(b'!H', bytes, 2, self.length_recovery)
-        struct.pack_into(b'!I', bytes, 4, self.mask)
-        bytes[4] = ((self.payload_type_recovery & FecPacket.PT_MASK) +
-                    (FecPacket.E_MASK if self.extended else 0))
-        struct.pack_into(b'!I', bytes, 8, self.timestamp_recovery)
-        bytes[12] = ((self.N_MASK if self.n else 0) +
-                     (self.D_MASK if self.direction else 0) +
-                     ((self.algorithm << FecPacket.T_SHIFT) & self.T_MASK) +
-                     (self.index & FecPacket.I_MASK))
-        bytes[13] = self.offset
-        bytes[14] = self.na
-        bytes[15] = self.snbase >> FecPacket.SNBE_SHIFT
-        return bytes
+        header = bytearray(self.HEADER_LENGTH)
+        struct.pack_into(b'!H', header, 0, self.snbase & self.SNBL_MASK)
+        struct.pack_into(b'!H', header, 2, self.length_recovery)
+        struct.pack_into(b'!I', header, 4, self.mask)
+        header[4] = (
+            (self.payload_type_recovery & self.PT_MASK)
+            + (self.E_MASK if self.extended else 0))
+        struct.pack_into(b'!I', header, 8, self.timestamp_recovery)
+        header[12] = (
+            (self.N_MASK if self.n else 0)
+            + (self.D_MASK if self.direction else 0)
+            + ((self.algorithm << self.T_SHIFT) & self.T_MASK)
+            + (self.index & self.I_MASK))
+        header[13] = self.offset
+        header[14] = self.na
+        header[15] = self.snbase >> self.SNBE_SHIFT
+        return header
 
     @property
     def bytes(self):
         return self.header_bytes + self.payload_recovery
 
-    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constructor >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constructor >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+
+    def __init__(self, data=None, length=0):
 
-    def __init__(self, bytes=None, length=0):
         # Fields default values
         self._errors = []
         self.sequence = 0
-        self.algorithm = FecPacket.XOR
-        self.direction = FecPacket.COL
+        self.algorithm = self.XOR
+        self.direction = self.COL
         self.snbase = 0
         self.offset = 0
-        self.na = 0
+        self.na = 0  # pylint:disable=invalid-name
         self.payload_type_recovery = 0
         self.timestamp_recovery = 0
         self.length_recovery = 0
         self.payload_recovery = []
-        # (Unused has defined in SMPTE 2022-1-1)
+        # (Unused as defined in SMPTE 2022-1-1)
         self.index = 0
         self.mask = 0
         self.extended = True
-        self.n = False
+        self.n = False  # pylint:disable=invalid-name
         self.missing = []
-        if bytes:
-            packet = RtpPacket(bytes, length)
+
+        if data is not None:
+            packet = RtpPacket(data, length)
             self.sequence = packet.sequence
-            self._errors = packet.errors
+            self._errors = packet.errors[:]
             if len(self._errors) > 0:
                 return
             if packet.payload_type != RtpPacket.DYNAMIC_PT:
-                self._errors.add(FecPacket.ER_PAYLOAD_TYPE)
+                self._errors.append(self.ER_PAYLOAD_TYPE)
                 return
-            self.snbase = (packet.payload[15]*256 + packet.payload[0])*256 + packet.payload[1]
-            self.length_recovery = packet.payload[2]*256 + packet.payload[3]
-            self.extended = (packet.payload[4] & FecPacket.E_MASK) != 0
-            #if not self.extended:
-            #    return
-            self.payload_type_recovery = packet.payload[4] & FecPacket.PT_MASK
-            self.mask = (packet.payload[5]*256 + packet.payload[6])*256 + packet.payload[7]
-            #if self.mask != 0:
-            #    return
-            self.timestamp_recovery = (((packet.payload[8]*256 + packet.payload[9])*256 +
-                                        packet.payload[10])*256 + packet.payload[11])
-            self.n = (packet.payload[12] & FecPacket.N_MASK) != 0
-            #if self.n:
-            #    return
-            self.direction = (packet.payload[12] & FecPacket.D_MASK) >> 6
-            self.algorithm = packet.payload[12] & FecPacket.T_MASK
-            # if self.algorithm != FecPacket.XOR:
-            #    return
-            self.index = packet.payload[12] & FecPacket.I_MASK
+            self.snbase = (packet.payload[15] * 256 + packet.payload[0]) * 256 + packet.payload[1]
+            self.length_recovery = packet.payload[2] * 256 + packet.payload[3]
+            self.extended = (packet.payload[4] & self.E_MASK) != 0
+            # if not self.extended:
+            #     return
+            self.payload_type_recovery = packet.payload[4] & self.PT_MASK
+            self.mask = (packet.payload[5] * 256 + packet.payload[6]) * 256 + packet.payload[7]
+            # if self.mask != 0:
+            #     return
+            self.timestamp_recovery = (
+                ((packet.payload[8] * 256 + packet.payload[9]) * 256 + packet.payload[10]) * 256
+                + packet.payload[11])
+            self.n = (packet.payload[12] & self.N_MASK) != 0
+            # if self.n:
+            #     return
+            self.direction = (packet.payload[12] & self.D_MASK) >> 6
+            self.algorithm = packet.payload[12] & self.T_MASK
+            # if self.algorithm != self.XOR:
+            #     return
+            self.index = packet.payload[12] & self.I_MASK
             # if self.index != 0:
-            #    return
+            #     return
             self.offset = packet.payload[13]
             self.na = packet.payload[14]
             self.snbase += packet.payload[15] << 16
             # And finally ... The payload !
-            self.payload_recovery = packet.payload[FecPacket.HEADER_LENGTH:]
+            self.payload_recovery = packet.payload[self.HEADER_LENGTH:]
 
-    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Functions >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Functions >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
 
-    @staticmethod
-    def compute(sequence, algorithm, direction, L, D, packets):
+    @classmethod
+    def compute(cls, sequence, algorithm, direction, L, D, packets):  # pylint:disable=invalid-name
         """
         This method will generate FEC packet's field by applying FEC algorithm to input packets.
-        In case of error (e.g. bad version number) the method will abort filling fields and un-updated fields are set to
-        their corresponding default value.
+        In case of error (e.g. bad version number) the method will abort filling fields and
+        un-updated fields are set to their corresponding default value.
 
         :param sequence: Sequence number of computed FEC packet
         :type sequence: int
@@ -416,9 +396,11 @@ class FecPacket(object):
 
         Testing invalid input collection of packets:
 
-        >>> from ..rtp import RtpPacket
-        >>> packets = [RtpPacket.create(10, 10, RtpPacket.MP2T_PT, 'a'),
-        ...            RtpPacket.create(22, 22, RtpPacket.MP2T_PT, 'b')]
+        >>> from pytoolbox.network.rtp import RtpPacket
+        >>> packets = [
+        ...     RtpPacket.create(10, 10, RtpPacket.MP2T_PT, 'a'),
+        ...     RtpPacket.create(22, 22, RtpPacket.MP2T_PT, 'b')
+        ... ]
         >>> fec = FecPacket.compute(1, FecPacket.XOR, FecPacket.COL, 2, 2, packets)
         Traceback (most recent call last):
             ...
@@ -426,10 +408,12 @@ class FecPacket(object):
 
         Testing valid input collection of packets:
 
-        >>> packets = [RtpPacket.create(10, 10, RtpPacket.MP2T_PT, bytearray('gaga', 'utf-8')),
-        ...            RtpPacket.create(14, 14, RtpPacket.MP2T_PT, bytearray('salut', 'utf-8')),
-        ...            RtpPacket.create(18, 18, RtpPacket.MP2T_PT, bytearray('12345', 'utf-8')),
-        ...            RtpPacket.create(22, 22, RtpPacket.MP2T_PT, bytearray('robot', 'utf-8'))]
+        >>> packets = [
+        ...     RtpPacket.create(10, 10, RtpPacket.MP2T_PT, bytearray('gaga', 'utf-8')),
+        ...     RtpPacket.create(14, 14, RtpPacket.MP2T_PT, bytearray('salut', 'utf-8')),
+        ...     RtpPacket.create(18, 18, RtpPacket.MP2T_PT, bytearray('12345', 'utf-8')),
+        ...     RtpPacket.create(22, 22, RtpPacket.MP2T_PT, bytearray('robot', 'utf-8'))
+        ... ]
         >>> fec = FecPacket.compute(2, FecPacket.XOR, FecPacket.COL, 4, 4, packets)
         >>> print(fec)
         errors                = []
@@ -452,99 +436,136 @@ class FecPacket(object):
 
         >>> from os import urandom
         >>> from random import randint
-        >>> from ..rtp import RtpPacket
+        >>> from pytoolbox.network.rtp import RtpPacket
         >>> L = 4
         >>> D = 5
         >>> OFF = 2
         >>> # Generate a [D][L] matrix of randomly generated RTP packets
-        >>> matrix = [[RtpPacket.create(L * j + i, (L * j + i) * 100 + randint(0, 50),
-        ...           RtpPacket.MP2T_PT, bytearray(urandom(randint(50, 100))))
-        ...           for i in xrange(L)] for j in xrange(D)]
-        >>> assert(len(matrix) == D and len(matrix[0]) == L)
-        >>> # Retrieve the OFF'th column of the matrix
+        >>> matrix = [
+        ...     [
+        ...         RtpPacket.create(L * j + i, (L * j + i) * 100 + randint(0, 50),
+        ...         RtpPacket.MP2T_PT, bytearray(urandom(randint(50, 100))))
+        ...         for i in range(L)
+        ...     ]
+        ...     for j in range(D)
+        ... ]
+        >>> len(matrix)  # D
+        5
+        >>> len(matrix[0])  # L
+        4
+
+        Retrieve the OFF'th column of the matrix:
+
         >>> expected_payload_type_recovery = 0
         >>> expected_timestamp_recovery = 0
         >>> expected_lenght_recovery = 0
         >>> expected_payload_recovery = bytearray(100)
         >>> packets = []
-        >>> for i in xrange(D):
+        >>> for i in range(D):
         ...     packet = matrix[i][OFF]
         ...     packets.append(packet)
         ...     # Compute expected recovery fields values
         ...     expected_payload_type_recovery ^= packet.payload_type
         ...     expected_timestamp_recovery ^= packet.timestamp
         ...     expected_lenght_recovery ^= packet.payload_size
-        ...     for j in xrange(packet.payload_size):
+        ...     for j in range(packet.payload_size):
         ...         expected_payload_recovery[j] ^= packet.payload[j]
         >>> fec = FecPacket.compute(15, FecPacket.XOR, FecPacket.COL, L, D, packets)
-        >>> assert(fec.valid)
-        >>> assert(fec.snbase == matrix[0][OFF].sequence == 2)
-        >>> assert(fec.na == D and fec.offset == L)
-        >>> assert(fec.payload_type_recovery == expected_payload_type_recovery)
-        >>> assert(fec.timestamp_recovery == expected_timestamp_recovery)
-        >>> assert(fec.length_recovery == expected_lenght_recovery)
-        >>> for i in xrange(fec.payload_size):
+        >>> fec.valid
+        True
+        >>> fec.snbase
+        2
+        >>> matrix[0][OFF].sequence
+        2
+        >>> fec.na  # D
+        5
+        >>> fec.offset  # L
+        4
+        >>> assert fec.payload_type_recovery == expected_payload_type_recovery
+        >>> assert fec.timestamp_recovery == expected_timestamp_recovery
+        >>> assert fec.length_recovery == expected_lenght_recovery
+        >>> for i in range(fec.payload_size):
         ...     if fec.payload_recovery[i] != expected_payload_recovery[i]:
         ...         print('Payload recovery test failed with i = ' + i)
         """
         # Fields default values
-        fec = FecPacket()
+        fec = cls()
         fec.sequence = sequence
-        if not algorithm in FecPacket.ALGORITHM_RANGE:
-            raise ValueError(to_bytes('algorithm is not a valid FEC algorithm'))
-        if not direction in FecPacket.DIRECTION_RANGE:
-            raise ValueError(to_bytes('direction is not a valid FEC direction'))
+        if algorithm not in cls.ALGORITHM_RANGE:
+            raise ValueError('algorithm is not a valid FEC algorithm')
+        if direction not in cls.DIRECTION_RANGE:
+            raise ValueError('direction is not a valid FEC direction')
         fec.algorithm = algorithm
         fec.direction = direction
-        if fec.direction == FecPacket.COL:
+        if fec.direction == cls.COL:
             fec.na = D
             fec.offset = L
         else:
             fec.na = L
             fec.offset = 1
-        if fec.algorithm != FecPacket.XOR:
-            raise NotImplementedError(to_bytes(FecPacket.ER_ALGORITHM))
+        if fec.algorithm != cls.XOR:
+            raise NotImplementedError(cls.ER_ALGORITHM)
         if len(packets) != fec.na:
-            raise ValueError(to_bytes('packets must contain exactly {0} packets'.format(fec.na)))
+            raise ValueError(f'packets must contain exactly {fec.na} packets')
         fec.snbase = packets[0].sequence
+
         # Detect maximum length of packets payload and check packets validity
-        size = 0
-        i = 0
+        size = i = 0
         for packet in packets:
             if not packet.validMP2T:
-                raise ValueError(to_bytes(FecPacket.ER_VALID_MP2T))
-            if packet.sequence != (fec.snbase + i*fec.offset) % RtpPacket.S_MASK:
-                raise ValueError(to_bytes(FecPacket.ER_SEQUENCE))
+                raise ValueError(cls.ER_VALID_MP2T)
+            if packet.sequence != (fec.snbase + i * fec.offset) & RtpPacket.S_MASK:
+                raise ValueError(cls.ER_SEQUENCE)
+
             size = max(size, packet.payload_size)
             i += 1
+
         # Create payload recovery field according to size/length
         fec.payload_recovery = bytearray(size)
+
         # Compute FEC packet's fields based on input packets
         for packet in packets:
+
             # Update (...) recovery fields by xor'ing corresponding fields of all packets
             fec.payload_type_recovery ^= packet.payload_type
             fec.timestamp_recovery ^= packet.timestamp
             fec.length_recovery ^= packet.payload_size
+
             # Update payload recovery by xor'ing all packets payload
             payload = packet.payload
             if len(packet.payload) < size:
                 payload = payload + bytearray(size - len(packet.payload))
+
             fast_xor_inplace(fec.payload_recovery, payload)
-            # NUMPY fec.payload_recovery = bytearray(numpy.bitwise_xor(fec.payload_recovery, payload))
+            # NUMPY fec.payload_recovery = bytearray(
+            #     numpy.bitwise_xor(fec.payload_recovery, payload))
             # XOR LOOP for i in xrange(min(size, len(packet.payload))):
             # XOR LOOP     fec.payload_recovery[i] ^= packet.payload[i]
         return fec
 
+    def compute_j(self, media_sequence):
+        """
+        TODO
+        """
+        delta = media_sequence - self.snbase
+        if delta < 0:
+            delta += RtpPacket.S_MASK + 1
+        if delta % self.offset != 0:
+            return None
+        return int(delta / self.offset)
+
     def set_missing(self, media_sequence):
         """
         Register a protected media packet as missing.
 
         **Example usage**
 
-        >>> packets = [RtpPacket.create(65530, 65530, RtpPacket.MP2T_PT, bytearray('gaga', 'utf-8')),
-        ...            RtpPacket.create(65533, 65533, RtpPacket.MP2T_PT, bytearray('salut', 'utf-8')),
-        ...            RtpPacket.create(    1,     1, RtpPacket.MP2T_PT, bytearray('12345', 'utf-8')),
-        ...            RtpPacket.create(    4,     4, RtpPacket.MP2T_PT, bytearray('robot', 'utf-8'))]
+        >>> packets = [
+        ...     RtpPacket.create(65530, 65530, RtpPacket.MP2T_PT, bytearray('gaga', 'utf-8')),
+        ...     RtpPacket.create(65533, 65533, RtpPacket.MP2T_PT, bytearray('salut', 'utf-8')),
+        ...     RtpPacket.create(    0,     0, RtpPacket.MP2T_PT, bytearray('12345', 'utf-8')),
+        ...     RtpPacket.create(    3,     3, RtpPacket.MP2T_PT, bytearray('robot', 'utf-8'))
+        ... ]
         >>> fec = FecPacket.compute(4, FecPacket.XOR, FecPacket.COL, 3, 4, packets)
         >>> print(fec)
         errors                = []
@@ -556,7 +577,7 @@ class FecPacket(object):
         na                    = 4
         L x D                 = 3 x 4
         payload type recovery = 0
-        timestamp recovery    = 2
+        timestamp recovery    = 4
         length recovery       = 1
         payload recovery size = 5
         missing               = []
@@ -566,58 +587,60 @@ class FecPacket(object):
         >>> fec.set_missing(fec.snbase + fec.offset + 1)
         Traceback (most recent call last):
             ...
-        ValueError: Unable to find a suitable j e N that satisfy : media_sequence = snbase + j * offset
+        ValueError: Unable to find a suitable j e N that satisfy : media_sequence = snbase + j * ...
         >>> fec.set_recovered(-1)
         Traceback (most recent call last):
             ...
-        ValueError: Unable to find a suitable j e N that satisfy : media_sequence = snbase + j * offset
+        ValueError: Unable to find a suitable j e N that satisfy : media_sequence = snbase + j * ...
 
         Testing set / get of a unique missing value:
 
-        >>> print(fec.set_missing(1))
+        >>> fec.set_missing(0)
         2
-        >>> print(fec.missing[0])
-        1
-        >>> print(len(fec.missing))
+        >>> fec.missing[0]
+        0
+        >>> len(fec.missing)
         1
 
         Testing simple recovery of a unique value:
 
-        >>> print(fec.set_recovered(1))
+        >>> fec.set_recovered(0)
         2
-        >>> print(len(fec.missing))
+        >>> len(fec.missing)
         0
 
         Testing set / get of multiple missing values (including re-setting of a value):
 
-        >>> print(fec.set_missing(4))
+        >>> fec.set_missing(3)
         3
-        >>> print(fec.set_missing(4))
+        >>> fec.set_missing(3)
         3
-        >>> print(len(fec.missing))
+        >>> len(fec.missing)
         1
-        >>> print(fec.set_missing(fec.snbase + fec.offset))
+        >>> fec.set_missing(fec.snbase + fec.offset)
         1
-        >>> print(fec.set_missing(fec.snbase))
+        >>> fec.set_missing(fec.snbase)
         0
-        >>> print(len(fec.missing))
+        >>> len(fec.missing)
         3
-        >>> print(fec.missing)
-        [4, 65533, 65530]
+        >>> fec.missing
+        [3, 65533, 65530]
 
         Testing re-recovery of a value:
 
-        >>> fec.set_recovered(4); fec.set_recovered(4)
+        >>> fec.set_recovered(3)
+        3
+        >>> fec.set_recovered(3)
         Traceback (most recent call last):
             ...
         ValueError: list.remove(x): x not in list
-        >>> print(fec.missing)
+        >>> fec.missing
         [65533, 65530]
         """
-        j = self.computeJ(media_sequence)
+        j = self.compute_j(media_sequence)
         if j is None:
-            raise ValueError(FecPacket.ER_J)
-        if not media_sequence in self.missing:
+            raise ValueError(self.ER_J)
+        if media_sequence not in self.missing:
             self.missing.append(media_sequence)
         return j
 
@@ -625,34 +648,24 @@ class FecPacket(object):
         """
         TODO
         """
-        j = self.computeJ(media_sequence)
+        j = self.compute_j(media_sequence)
         if j is None:
-            raise ValueError(FecPacket.ER_J)
+            raise ValueError(self.ER_J)
         self.missing.remove(media_sequence)
         return j
 
-    def computeJ(self, media_sequence):
-        """
-        TODO
-        """
-        delta = media_sequence - self.snbase
-        if delta < 0:
-            delta = RtpPacket.S_MASK + delta
-        if delta % self.offset != 0:
-            return None
-        return int(delta / self.offset)
-
     def __eq__(self, other):
         """
         Test equality field by field !
         """
-        if isinstance(other, self.__class__):
-            return (self.sequence == other.sequence and self.algorithm == other.algorithm and
-                    self.direction == other.direction and self.snbase == other.snbase and
-                    self.offset == other.offset and self.na == other.na and
-                    self.payload_type_recovery == other.payload_type_recovery and
-                    self.length_recovery == other.length_recovery and
-                    self.payload_recovery == other.payload_recovery)
+        return (
+            isinstance(other, self.__class__)
+            and self.sequence == other.sequence and self.algorithm == other.algorithm
+            and self.direction == other.direction and self.snbase == other.snbase
+            and self.offset == other.offset and self.na == other.na
+            and self.payload_type_recovery == other.payload_type_recovery
+            and self.length_recovery == other.length_recovery
+            and self.payload_recovery == other.payload_recovery)
 
     def __str__(self):
         """
@@ -696,17 +709,16 @@ class FecPacket(object):
         payload recovery size = 10
         missing               = []
         """
-        return ("""errors                = {0.errors}
-sequence              = {0.sequence}
-algorithm             = {1}
-direction             = {2}
-snbase                = {0.snbase}
-offset                = {0.offset}
-na                    = {0.na}
-L x D                 = {0.L} x {0.D}
-payload type recovery = {0.payload_type_recovery}
-timestamp recovery    = {0.timestamp_recovery}
-length recovery       = {0.length_recovery}
-payload recovery size = {0.payload_size}
-missing               = {0.missing}""".format(self, FecPacket.ALGORITHM_NAMES[self.algorithm],
-                                              FecPacket.DIRECTION_NAMES[self.direction]))
+        return f"""errors                = {self.errors}
+sequence              = {self.sequence}
+algorithm             = {self.ALGORITHM_NAMES[self.algorithm]}
+direction             = {self.DIRECTION_NAMES[self.direction]}
+snbase                = {self.snbase}
+offset                = {self.offset}
+na                    = {self.na}
+L x D                 = {self.L} x {self.D}
+payload type recovery = {self.payload_type_recovery}
+timestamp recovery    = {self.timestamp_recovery}
+length recovery       = {self.length_recovery}
+payload recovery size = {self.payload_size}
+missing               = {self.missing}"""
