@@ -1,102 +1,79 @@
-# -*- encoding: utf-8 -*-
-
-#**********************************************************************************************************************#
-#                                        PYTOOLBOX - TOOLBOX FOR PYTHON SCRIPTS
-#
-#  Main Developer : David Fischer (david.fischer.ch@gmail.com)
-#  Copyright      : Copyright (c) 2012-2014 David Fischer. All rights reserved.
-#
-#**********************************************************************************************************************#
-#
-# This file is part of David Fischer's pytoolbox Project.
-#
-# This project is free software: you can redistribute it and/or modify it under the terms of the EUPL v. 1.1 as provided
-# by the European Commission. This project is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
-# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-#
-# See the European Union Public License for more details.
-#
-# You should have received a copy of the EUPL General Public License along with this project.
-# If not, see he EUPL licence v1.1 is available in 22 languages:
-#     22-07-2013, <https://joinup.ec.europa.eu/software/page/eupl/licence-eupl>
-#
-# Retrieved from https://github.com/davidfischer-ch/pytoolbox.git
-
-from __future__ import absolute_import, division, print_function, unicode_literals
-
-from collections import defaultdict
+import collections, os
 
+from pytoolbox.network.ip import IPSocket
+from pytoolbox.network.rtp import RtpPacket
 from .base import FecPacket
-from ..ip import IPSocket
-from ..rtp import RtpPacket
-from ...encoding import to_bytes
 
-__all__ = ('FecReceiver', )
+__all__ = ['FecReceiver']
 
 
-class FecReceiver(object):
+class FecReceiver(object):  # pylint:disable=too-many-instance-attributes
     """
-    TODO
+    A SMPTE 2022-1 FEC streams receiver.
+    This receiver accept incoming RTP media and FEC packets and make available the recovered media
+    stream.
 
     **Example usage (with a network capture)**
 
-    >> import dpkt
-    >> from pytoolbox.network.smpte2022.base import FecPacket, RtpPacket
-    >> from pytoolbox.network.smpte2022.receiver import FecReceiver
-    >> from struct import *
-    >>
-    >> fec_receiver = FecReceiver(open("test.ts", "wb"))
-    >> fec_receiver.set_delay(1024, FecReceiver.PACKETS)
-    >>
-    >> for ts, buf in dpkt.pcap.Reader(open('test.dump')):
-    >>     udp = dpkt.ethernet.Ethernet(buf).data.data
-    >>     if udp.dport == 3300:
-    >>         media_packet = RtpPacket.create(unpack('!h', udp.data[2:4])[0], unpack('!i', udp.data[4:8])[0],
-    >>                                         RtpPacket.MP2T_PT, bytearray(udp.data[12:]))
-    >>         fec_receiver.put_media(media_packet, onlyMP2TS=True)
-    >>     elif udp.dport in (3302, 3304):
-    >>         fec_data = bytearray(udp.data)
-    >>         fec_receiver.put_fec(FecPacket(fec_data, len(fec_data)))
-    >> print(fec_receiver)
-    >> fec_receiver.flush()
+    ::
+
+        >> import dpkt
+        >> from pytoolbox.network.smpte2022.base import FecPacket, RtpPacket
+        >> from pytoolbox.network.smpte2022.receiver import FecReceiver
+        >> from struct import *
+        >>
+        >> fec_receiver = FecReceiver(open('test.ts', 'wb'))
+        >> fec_receiver.set_delay(1024, FecReceiver.PACKETS)
+        >>
+        >> for ts, buf in dpkt.pcap.Reader(open('test.dump')):
+        >>     udp = dpkt.ethernet.Ethernet(buf).data.data
+        >>     if udp.dport == 3300:
+        >>         media_packet = RtpPacket.create(
+        ..             unpack('!h', udp.data[2:4])[0], unpack('!i', udp.data[4:8])[0],
+        ..             RtpPacket.MP2T_PT, bytearray(udp.data[12:]))
+        >>         fec_receiver.put_media(media_packet, onlyMP2TS=True)
+        >>     elif udp.dport in (3302, 3304):
+        >>         fec_data = bytearray(udp.data)
+        >>         fec_receiver.put_fec(FecPacket(fec_data, len(fec_data)))
+        >> print(fec_receiver)
+        >> fec_receiver.flush()
 
     **Example usage**
 
     Media packets are sorted by the buffer, so, it's time to test this feature:
 
-    >>> import random
-    >>> from StringIO import StringIO
-    >>> output = StringIO()
+    >>> import io, random
+    >>> output = io.StringIO()
     >>> receiver = FecReceiver(output)
-    >>> receiver.set_delay(1024, FecReceiver.PACKETS)
-    >>> source = range(1024)
+    >>> receiver.set_delay(1000, FecReceiver.PACKETS)
+    >>> source = list(range(30))
     >>> random.shuffle(source)
     >>> for i in source:
     ...     receiver.put_media(RtpPacket.create(i, i * 100, RtpPacket.MP2T_PT, str(i)), True)
     >>> receiver.flush()
-    >>> assert(output.getvalue() == ''.join(str(i) for i in xrange(1024)))
+    >>> output.getvalue()
+    '01234567891011121314151617181920212223242526272829'
 
     Testing FEC algorithm correctness:
 
-    >>> from io import BytesIO
-    >>> from os import urandom
-    >>> from random import randint
-    >>> output = BytesIO()
+    >>> import io, os, random
+    >>> output = io.BytesIO()
     >>> receiver = FecReceiver(output)
     >>> receiver.set_delay(1024, FecReceiver.PACKETS)
     >>> L, D = 4, 5
     >>> # Generate a [D][L] matrix of randomly generated RTP packets
-    >>> matrix = [[RtpPacket.create(L * j + i, (L * j + i) * 100 + randint(0, 50),
-    ...           RtpPacket.MP2T_PT, bytearray(urandom(randint(50, 100))))
-    ...           for i in xrange(L)] for j in xrange(D)]
-    >>> assert(len(matrix) == D and len(matrix[0]) == L)
+    >>> matrix = [[RtpPacket.create(L * j + i, (L * j + i) * 100 + random.randint(0, 50),
+    ...           RtpPacket.MP2T_PT, bytearray(os.urandom(random.randint(50, 100))))
+    ...           for i in range(L)] for j in range(D)]
+    >>> assert len(matrix) == D and len(matrix[0]) == L
     >>> # Retrieve the first column of the matrix
     >>> for column in matrix:
     ...     for media in column:
     ...         if media.sequence != 0:
     ...             receiver.put_media(media, True)
     >>> fec = FecPacket.compute(1, FecPacket.XOR, FecPacket.COL, L, D, [p[0] for p in matrix[0:]])
-    >>> print('dir={0} snbase={1} offset={2} na={3}'.format(fec.direction, fec.snbase, fec.offset, fec.na))
+    >>> print('dir={0} snbase={1} offset={2} na={3}'.format(
+    ...     fec.direction, fec.snbase, fec.offset, fec.na))
     dir=0 snbase=0 offset=4 na=5
     >>> receiver.put_fec(fec)
     >>> print(receiver)
@@ -121,18 +98,22 @@ class FecReceiver(object):
     True
     """
 
-    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constants >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constants >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
 
     ER_DELAY_UNITS = "Unknown delay units '{0}'"
     ER_DIRECTION = 'FEC packet direction is neither COL nor ROW : {0}'
     ER_FLUSHING = 'Currently flushing buffers'
     ER_MISSING_COUNT = 'They are {0} missing media packet, expected one (1)'
+    ER_FEC_DIRECTION = "Invalid FEC direction '{0}'"
     ER_COL_MISMATCH = 'Column FEC packet n°{0}, expected n°{1}'
-    ER_COL_OVERWRITE = 'Another column FEC packet is already registered to protect media packet n°{0}'
+    ER_COL_OVERWRITE = \
+        'Another column FEC packet is already registered to protect media packet n°{0}'
     ER_ROW_MISMATCH = 'Row FEC packet n°{0}, expected n°{1}'
     ER_ROW_OVERWRITE = 'Another row FEC packet is already registered to protect media packet n°{0}'
-    ER_GET_COL_CASCADE = 'Column FEC cascade : Unable to compute sequence # of the media packet to recover\n{0}\n'
-    ER_GET_ROW_CASCADE = 'Row FEC cascade : Unable to compute sequence # of the media packet to recover\n{0}\n'
+    ER_GET_COL_CASCADE = \
+        'Column FEC cascade : Unable to compute sequence # of the media packet to recover{0}{1}{0}'
+    ER_GET_ROW_CASCADE = \
+        'Row FEC cascade : Unable to compute sequence # of the media packet to recover{0}{1}{0}'
     ER_NULL_COL_CASCADE = 'Column FEC cascade : Unable to find linked entry in crosses buffer'
     ER_NULL_ROW_CASCADE = 'Row FEC cascade : Unable to find linked entry in crosses buffer'
     ER_STARTUP = 'Current position still not initialized (startup state)'
@@ -140,14 +121,14 @@ class FecReceiver(object):
     ER_VALID_RTP = 'packet is not valid (expected RTP packet)'
 
     DELAY_NAMES = ['packets', 'seconds']
-    DELAY_RANGE = xrange(len(DELAY_NAMES))
+    DELAY_RANGE = range(len(DELAY_NAMES))  # noqa
     PACKETS, SECONDS = DELAY_RANGE
 
-    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constructors >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constructors >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
 
     def __init__(self, output):
         """
-        Construct a new FecReceiver and register ``output``.
+        Construct a new FecReceiver and register `output`.
 
         :param output: Where to output payload of the recovered stream.
         :type output: IOBase
@@ -161,12 +142,12 @@ class FecReceiver(object):
             ...
         ValueError: output is None
         >>>
-        >>> from StringIO import StringIO
-        >>> output = StringIO()
+        >>> import io
+        >>> output = io.StringIO()
         >>> receiver = FecReceiver(output)
         """
         if not output:
-            raise ValueError(to_bytes('output is None'))
+            raise ValueError('output is None')
         # Media packets storage, medias[media seq] = media pkt
         self.medias = {}
         self.startup = True    # Indicate that actual position must be initialized
@@ -174,16 +155,16 @@ class FecReceiver(object):
         self.position = 0      # Actual position (sequence number) in the medias buffer
         # Link media packets to fec packets able to recover it, crosses[mediaseq] = {colseq, rowseq}
         self.crosses = {}
-        # Fec packets + related informations storage, col[sequence] = { fec pkt + infos }
+        # Fec packets + related information storage, col[sequence] = { fec pkt + info }
         self.cols = {}
         self.rows = {}
-        self.matrixL = 0  # Detected FEC matrix size (number of columns)
-        self.matrixD = 0  # Detected FEC matrix size (number of rows)
+        self.matrixL = 0  # Detected FEC matrix size (number of columns) pylint:disable=invalid-name
+        self.matrixD = 0  # Detected FEC matrix size (number of rows)    pylint:disable=invalid-name
         # Output
         self.output = output  # Registered output
         # Settings
-        self.delay_value = 100                  # RTP buffer delay value
-        self.delay_units = FecReceiver.PACKETS  # RTP buffer delay units
+        self.delay_value = 100           # RTP buffer delay value
+        self.delay_units = self.PACKETS  # RTP buffer delay units
         # Statistics about media (buffers and packets)
         self.media_received = 0          # Received media packets counter
         self.media_recovered = 0         # Recovered media packets counter
@@ -199,45 +180,45 @@ class FecReceiver(object):
         self.max_cross = 0     # Largest amount of stored elements in the crosses buffer
         self.max_col = 0       # Largest amount of stored elements in the columns buffer
         self.max_row = 0       # Largest amount of stored elements in the rows buffer
-        self.lostogram = defaultdict(int)  # Statistics about lost medias
-        self.lostogram_counter = 0         # Incremented while there are lost media packets
+        self.lostogram = collections.defaultdict(int)  # Statistics about lost medias
+        self.lostogram_counter = 0  # Incremented while there are lost media packets
 
-    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Properties >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Properties >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
 
     @property
     def current_delay(self):
         """Return current delay based on the length of the media buffer."""
         if len(self.medias) == 0:
             return 0
-        if self.delay_units == FecReceiver.PACKETS:
+        if self.delay_units == self.PACKETS:
             return len(self.medias)
-        elif self.delay_units == FecReceiver.SECONDS:
+        if self.delay_units == self.SECONDS:
             raise NotImplementedError()
-        raise ValueError(to_bytes(FecReceiver.ER_DELAY_UNITS.format(self.delay_units)))
-            #return medias.lastEntry().getValue().getTime() - medias.firstEntry().getValue().getTime()
+        raise ValueError(self.ER_DELAY_UNITS.format(self.delay_units))
+        # return medias.lastEntry().getValue().getTime() - medias.firstEntry().getValue().getTime()
 
-    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Functions >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Functions >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
 
     def set_delay(self, value, units):
         """Set desired size for the internal media buffer."""
-        if units == FecReceiver.PACKETS:
+        if units == self.PACKETS:
             self.delay_value = value
             self.delay_units = units
-        elif units == FecReceiver.SECONDS:
+        elif units == self.SECONDS:
             raise NotImplementedError()
         else:
-            raise ValueError(to_bytes(FecReceiver.ER_DELAY_UNITS.format(units)))
+            raise ValueError(self.ER_DELAY_UNITS.format(units))
 
-    def put_media(self, media, onlyMP2TS):
+    def put_media(self, media, onlyMP2TS):  # pylint:disable=invalid-name
         """Put an incoming media packet."""
         if self.flushing:
-            raise ValueError(to_bytes(FecReceiver.ER_FLUSHING))
+            raise ValueError(self.ER_FLUSHING)
         if onlyMP2TS:
             if not media.validMP2T:
-                raise ValueError(to_bytes(FecReceiver.ER_VALID_RTP_MP2TS))
+                raise ValueError(self.ER_VALID_RTP_MP2TS)
         else:
             if not media.valid:
-                raise ValueError(to_bytes(FecReceiver.ER_VALID_RTP))
+                raise ValueError(self.ER_VALID_RTP)
 
         # Put the media packet into medias buffer
         if media.sequence in self.medias:
@@ -254,32 +235,35 @@ class FecReceiver(object):
 
         self.out()  # FIXME maybe better to call it from another thread
 
-    def put_fec(self, fec):
+    def put_fec(self, fec):  # pylint:disable=too-many-branches,too-many-statements
         """
-        Put an incoming FEC packet, the algorithm will do the following according to these scenarios:
+        Put an incoming FEC packet, the algorithm will do the following according to these
+        scenarios:
 
         1. The fec packet is useless if none of the protected media packets is missing
         2. Only on media packet missing, fec packet is able to recover it now !
         3. More than one media packet is missing, fec packet stored for future recovery
         """
         if self.flushing:
-            raise ValueError(to_bytes(FecReceiver.ER_FLUSHING))
+            raise ValueError(self.ER_FLUSHING)
         if not fec.valid:
-            raise ValueError(to_bytes('Invalid FEC packet'))
+            raise ValueError('Invalid FEC packet')
 
         if fec.direction == FecPacket.COL:
             self.col_received += 1
         elif fec.direction == FecPacket.ROW:
             self.row_received += 1
         else:
-            raise ValueError(FecReceiver.ER_DIRECTION.format(fec.direction))
+            raise ValueError(self.ER_DIRECTION.format(fec.direction))
+
         cross = None
         media_lost = 0
         media_max = (fec.snbase + fec.na * fec.offset) & RtpPacket.S_MASK
         media_test = fec.snbase
+
         while media_test != media_max:
             # If media packet is not in the medias buffer (is missing)
-            if not media_test in self.medias:
+            if media_test not in self.medias:
                 media_lost = media_test
                 # TODO
                 cross = self.crosses.get(media_test)
@@ -288,30 +272,43 @@ class FecReceiver(object):
                     self.crosses[media_test] = cross
                     if len(self.crosses) > self.max_cross:
                         self.max_cross = len(self.crosses)
+
                 # Register the fec packet able to recover the missing media packet
                 if fec.direction == FecPacket.COL:
                     if cross['col_sequence']:
-                        raise ValueError(to_bytes(FecReceiver.ER_COL_OVERWRITE.format(media_lost)))
+                        raise ValueError(self.ER_COL_OVERWRITE.format(media_lost))
                     cross['col_sequence'] = fec.sequence
+
                 elif fec.direction == FecPacket.ROW:
                     if cross['row_sequence']:
-                        raise ValueError(to_bytes(FecReceiver.ER_ROW_OVERWRITE.format(media_lost)))
+                        raise ValueError(self.ER_ROW_OVERWRITE.format(media_lost))
                     cross['row_sequence'] = fec.sequence
+
                 else:
-                    raise ValueError(to_bytes(FecReceiver.ER_FEC_DIRECTION.format(fec.direction)))
+                    raise ValueError(self.ER_FEC_DIRECTION.format(fec.direction))
+
                 fec.set_missing(media_test)
+
             media_test = (media_test + fec.offset) & RtpPacket.S_MASK
-        if fec.L != 0:
+
+        if fec.L:
             self.matrixL = fec.L
-        if fec.D != 0:
+
+        if fec.D:
             self.matrixD = fec.D
+
         # [1] The fec packet is useless if none of the protected media packets is missing
         if len(fec.missing) == 0:
             return
+
         # FIXME check if 10*delay_value is a good way to avoid removing early fec packets !
         # The fec packet is useless if it needs an already output'ed media packet to do recovery
-        drop = not FecReceiver.validity_window(fec.snbase, self.position,
-                                              (self.position + 10 * self.delay_value) & RtpPacket.S_MASK)
+        drop = not self.validity_window(
+            fec.snbase,
+            self.position,
+            (self.position + 10 * self.delay_value) & RtpPacket.S_MASK
+        )
+
         if fec.direction == FecPacket.COL:
             if drop:
                 self.col_dropped += 1
@@ -319,6 +316,7 @@ class FecReceiver(object):
             self.cols[fec.sequence] = fec
             if len(self.cols) > self.max_col:
                 self.max_col = len(self.cols)
+
         if fec.direction == FecPacket.ROW:
             if drop:
                 self.row_dropped += 1
@@ -326,6 +324,7 @@ class FecReceiver(object):
             self.rows[fec.sequence] = fec
             if len(self.rows) > self.max_row:
                 self.max_row = len(self.rows)
+
         # [2] Only on media packet missing, fec packet is able to recover it now !
         if len(fec.missing) == 1:
             self.recover_media_packet(media_lost, cross, fec)
@@ -344,26 +343,34 @@ class FecReceiver(object):
     def cleanup(self):
         """Remove FEC packets that are stored / waiting but useless."""
         if self.flushing:
-            raise ValueError(to_bytes(FecReceiver.ER_FLUSHING))
+            raise ValueError(self.ER_FLUSHING)
+
         if self.startup:
-            raise ValueError(to_bytes(FecReceiver.ER_STARTUP))
-        if self.delay_units == FecReceiver.PACKETS:
+            raise ValueError(self.ER_STARTUP)
+
+        if self.delay_units == self.PACKETS:
             start, end = self.position, (self.position + self.delay_value) & RtpPacket.S_MASK
-            for media_sequence in self.crosses.keys:
+            for media_sequence, cross in list(self.crosses.items()):
                 if not self.validity_window(media_sequence, start, end):
-                    cross = self.crosses[media_sequence]
                     del self.cols[cross['col_sequence']]
                     del self.rows[cross['row_sequence']]
                     del self.crosses[media_sequence]
-        elif self.delay_units == FecReceiver.SECONDS:
+        elif self.delay_units == self.SECONDS:
             raise NotImplementedError()
-        raise ValueError(to_bytes(FecReceiver.ER_DELAY_UNITS.format(self.delay_units)))
 
-    def recover_media_packet(self, media_sequence, cross, fec):
-        """Recover a missing media packet helped by a FEC packet, this method is also called to register an incoming
-        media packet if it is registered as missing."""
+        raise ValueError(self.ER_DELAY_UNITS.format(self.delay_units))
 
-        recovered_by_fec = not fec is None
+    def recover_media_packet(
+        self,
+        media_sequence,
+        cross,
+        fec
+    ):  # pylint:disable=too-many-branches,too-many-locals,too-many-statements
+        """
+        Recover a missing media packet helped by a FEC packet, this method is also called to
+        register an incoming media packet if it is registered as missing.
+        """
+        recovered_by_fec = fec is not None
 
         # Read and remove "cross" it from the buffer
         col_sequence = cross['col_sequence']
@@ -372,17 +379,23 @@ class FecReceiver(object):
 
         # Recover the missing media packet and remove any useless linked fec packet
         if recovered_by_fec:
+
             if len(fec.missing) != 1:
-                raise NotImplementedError(FecReceiver.ER_MISSING_COUNT.format(len(fec.missing)))
+                raise NotImplementedError(self.ER_MISSING_COUNT.format(len(fec.missing)))
+
             if fec.direction == FecPacket.COL and fec.sequence != col_sequence:
-                raise NotImplementedError(FecReceiver.ER_COL_MISMATCH.format(fec.sequence, col_sequence))
+                raise NotImplementedError(self.ER_COL_MISMATCH.format(fec.sequence, col_sequence))
+
             if fec.direction == FecPacket.ROW and fec.sequence != row_sequence:
-                raise NotImplementedError(FecReceiver.ER_ROW_MISMATCH.format(fec.sequence, row_sequence))
+                raise NotImplementedError(self.ER_ROW_MISMATCH.format(fec.sequence, row_sequence))
 
             # Media packet recovery
             # > Copy fec packet fields into the media packet
-            media = RtpPacket.create(media_sequence, fec.timestamp_recovery, fec.payload_type_recovery,
-                                     fec.payload_recovery)
+            media = RtpPacket.create(
+                media_sequence,
+                fec.timestamp_recovery,
+                fec.payload_type_recovery,
+                fec.payload_recovery)
             payload_size = fec.length_recovery
 
             # > recovered payload ^= all media packets linked to the fec packet
@@ -403,8 +416,8 @@ class FecReceiver(object):
                 media.timestamp ^= friend.timestamp
                 payload_size ^= friend.payload_size
                 # FIXME FIXME FIXME FIXME FIXME OPTIMIZATION FIXME FIXME FIXME FIXME
-                for no in xrange(min(len(media.payload), len(friend.payload))):
-                    media.payload[no] ^= friend.payload[no]
+                for index in range(min(len(media.payload), len(friend.payload))):
+                    media.payload[index] ^= friend.payload[index]
                 media_test = (media_test + fec.offset) & RtpPacket.S_MASK
 
             # If the media packet is successfully recovered
@@ -440,11 +453,12 @@ class FecReceiver(object):
                         self.recover_media_packet(cascade_media_sequence, cascade_cross, fec_col)
                     else:
                         raise NotImplementedError(
-                            to_bytes('recover_media_packet({0}, {1}, {2}):\n{3}\nmedia sequence : {4}\n{5}\n'.format(
-                                     media_sequence, cross, fec, FecReceiver.ER_NULL_COL_CASCADE,
-                                     cascade_media_sequence, fec_col)))
+                            f'recover_media_packet({media_sequence}, {cross}, {fec}):{os.linesep}'
+                            f'{self.ER_NULL_COL_CASCADE}{os.linesep}'
+                            f'media sequence : {cascade_media_sequence}{os.linesep}'
+                            f'{fec_col}{os.linesep}')
                 else:
-                    raise NotImplementedError(to_bytes(FecReceiver.ER_GET_COL_CASCADE.format(fec_col)))
+                    raise NotImplementedError(self.ER_GET_COL_CASCADE.format(os.linesep, fec_col))
 
         if fec_row:
             if len(fec_row.missing) == 1:
@@ -456,24 +470,30 @@ class FecReceiver(object):
                         self.recover_media_packet(cascade_media_sequence, cascade_cross, fec_row)
                     else:
                         raise NotImplementedError(
-                            to_bytes('{0}\nrecover_media_packet({1}, {2}, {3}):\nmedia sequence : {4}\n{5}\n'.format(
-                                     FecReceiver.ER_NULL_ROW_CASCADE, media_sequence, cross, fec,
-                                     cascade_media_sequence, fec_row)))
+                            f'{self.ER_NULL_ROW_CASCADE}{os.linesep}'
+                            f'recover_media_packet({media_sequence}, {cross}, {fec}):{os.linesep}'
+                            f'media sequence : {cascade_media_sequence}{os.linesep}'
+                            f'{fec_row}{os.linesep}')
                 else:
-                    raise NotImplementedError(to_bytes(FecReceiver.ER_GET_ROW_CASCADE.format(fec_row)))
+                    raise NotImplementedError(self.ER_GET_ROW_CASCADE.format(os.linesep, fec_row))
 
     def out(self):
         """Extract packets to output in order to keep a 'certain' amount of them in the buffer."""
-        units = FecReceiver.PACKETS if self.flushing else self.delay_units
+        units = self.PACKETS if self.flushing else self.delay_units
         value = 0 if self.flushing else self.delay_value
         lostogram_counter = 0
+
         # Extract packets to output in order to keep a 'certain' amount of them in the buffer
-        if units == FecReceiver.PACKETS:  # based on buffer size
+        if units == self.PACKETS:  # based on buffer size
             while len(self.medias) > value:
                 # Initialize or increment actual position (expected sequence number)
-                self.position = ((self.medias.iterkeys().next() if self.startup else (self.position + 1))
-                                 & RtpPacket.S_MASK)
+                if self.startup:
+                    self.position = min(self.medias.keys()) & RtpPacket.S_MASK
+                else:
+                    self.position = (self.position + 1) & RtpPacket.S_MASK
+
                 self.startup = False
+
                 media = self.medias.get(self.position)
                 if media:
                     self.lostogram[lostogram_counter] += 1
@@ -484,19 +504,21 @@ class FecReceiver(object):
                 else:
                     self.media_missing += 1
                     lostogram_counter += 1
+
                 # Remove any fec packet linked to current media packet
                 cross = self.crosses.get(self.position)
                 if cross:
                     del self.crosses[self.position]
-                    if cross['col_sequence']:
+                    if cross['col_sequence'] and cross['col_sequence'] in self.cols:
                         del self.cols[cross['col_sequence']]
-                    if cross['row_sequence']:
+                    if cross['row_sequence'] and cross['row_sequence'] in self.rows:
                         del self.rows[cross['row_sequence']]
+
         # Extract packets to output in order to keep a 'certain' amount of them in the buffer
-        elif units == FecReceiver.SECONDS:  # based on time stamps
+        elif units == self.SECONDS:  # based on time stamps
             raise NotImplementedError()
         else:
-            raise ValueError(FecReceiver.ER_DELAY_UNITS.format(units))
+            raise ValueError(self.ER_DELAY_UNITS.format(units))
 
     def __str__(self):
         """
@@ -517,25 +539,26 @@ class FecReceiver(object):
         Current delay (can be set) : 0 packets
         FEC matrix size (LxD) : 0x0 = 0 packets
         """
-        delayFormat = '%.0f' if self.delay_units == FecReceiver.PACKETS else '%.2f'
-        mDelay = (delayFormat % self.current_delay) + ' ' + FecReceiver.DELAY_NAMES[self.delay_units]
-        return ("Name  Received Buffered Maximum Dropped\n"
-                "Media %8d%9d%8d\n"
-                "Col   %8d%9d%8d%8d\n"
-                "Row   %8d%9d%8d%8d\n"
-                "Cross         %9d%8d\n"
-                "FEC statistics, media packets :\n"
-                "Recovered Aborted Overwritten Missing\n"
-                "%9d%8d%12d%8d\n"
-                "Current position (media sequence) : %s\n"
-                "Current delay (can be set) : %s\n"
-                "FEC matrix size (LxD) : %sx%s = %s packets" %
+        delay_format = '%.0f' if self.delay_units == self.PACKETS else '%.2f'
+        delay_msg = f'{delay_format % self.current_delay} {self.DELAY_NAMES[self.delay_units]}'
+
+        return ("Name  Received Buffered Maximum Dropped{0}"
+                "Media %8d%9d%8d{0}"
+                "Col   %8d%9d%8d%8d{0}"
+                "Row   %8d%9d%8d%8d{0}"
+                "Cross         %9d%8d{0}"
+                "FEC statistics, media packets :{0}"
+                "Recovered Aborted Overwritten Missing{0}"
+                "%9d%8d%12d%8d{0}"
+                "Current position (media sequence) : %s{0}"
+                "Current delay (can be set) : %s{0}"
+                "FEC matrix size (LxD) : %sx%s = %s packets".format(os.linesep) %
                 (self.media_received, len(self.medias), self.max_media,
                  self.col_received, len(self.cols), self.max_col, self.col_dropped,
                  self.row_received, len(self.rows), self.max_row, self.row_dropped,
                  len(self.crosses), self.max_cross, self.media_recovered,
                  self.media_aborted_recovery, self.media_overwritten, self.media_missing,
-                 self.position, mDelay, self.matrixL, self.matrixD, self.matrixL * self.matrixD))
+                 self.position, delay_msg, self.matrixL, self.matrixD, self.matrixL * self.matrixD))
 
     # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Static >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
 
@@ -546,13 +569,17 @@ class FecReceiver(object):
 
         **Example usage**
 
-        >>> from nose.tools import eq_
-        >>> eq_(FecReceiver.compute_col_address('192.168.50.100:8000'), {'ip': '192.168.50.100', 'port': 8002})
-        >>> eq_(FecReceiver.compute_col_address(IPSocket('50.0.0.7:4000')), {'ip': '50.0.0.7', 'port': 4002})
+        >>> from pytoolbox.unittest import asserts
+        >>> asserts.dict_equal(FecReceiver.compute_col_address('192.168.50.100:8000'), {
+        ...     'ip': '192.168.50.100', 'port': 8002
+        ... })
+        >>> asserts.dict_equal(FecReceiver.compute_col_address(IPSocket('50.0.0.7:4000')), {
+        ...     'ip': '50.0.0.7', 'port': 4002
+        ... })
         >>> print(FecReceiver.compute_col_address('salut'))
         Traceback (most recent call last):
             ....
-        ValueError: salut is not a valid IP socket.
+        pytoolbox.exceptions.InvalidIPSocketError: salut is not a valid IP socket.
         """
         if not isinstance(media_socket, dict):
             media_socket = IPSocket(media_socket)
@@ -566,13 +593,17 @@ class FecReceiver(object):
 
         **Example usage**
 
-        >>> from nose.tools import eq_
-        >>> eq_(FecReceiver.compute_row_address('192.168.50.100:8000'), {'ip': '192.168.50.100', 'port': 8004})
-        >>> eq_(FecReceiver.compute_row_address(IPSocket('50.0.0.7:4000')), {'ip': '50.0.0.7', 'port': 4004})
+        >>> from pytoolbox.unittest import asserts
+        >>> asserts.dict_equal(FecReceiver.compute_row_address('192.168.50.100:8000'), {
+        ...     'ip': '192.168.50.100', 'port': 8004
+        ... })
+        >>> asserts.dict_equal(FecReceiver.compute_row_address(IPSocket('50.0.0.7:4000')), {
+        ...     'ip': '50.0.0.7', 'port': 4004
+        ... })
         >>> print(FecReceiver.compute_row_address('salut'))
         Traceback (most recent call last):
             ....
-        ValueError: salut is not a valid IP socket.
+        pytoolbox.exceptions.InvalidIPSocketError: salut is not a valid IP socket.
         """
         if not isinstance(media_socket, dict):
             media_socket = IPSocket(media_socket)
@@ -582,7 +613,7 @@ class FecReceiver(object):
     @staticmethod
     def validity_window(current, start, end):
         """
-        Returns True if ``current`` is in the validity window bounded by ``start`` and ``end``.
+        Returns True if `current` is in the validity window bounded by `start` and `end`.
 
         This method is circular-buffer aware and they are 2 cases (validity window ``[====]``)::
 
@@ -593,18 +624,27 @@ class FecReceiver(object):
 
         Testing validity window condition:
 
-        >>> assert(not FecReceiver.validity_window(    0,     5, 10))
-        >>> assert(    FecReceiver.validity_window(    5,     5, 10))
-        >>> assert(    FecReceiver.validity_window(    8,     5, 10))
-        >>> assert(    FecReceiver.validity_window(   10,     5, 10))
-        >>> assert(not FecReceiver.validity_window(   15,     5, 10))
-        >>> assert(    FecReceiver.validity_window(    0, 65534,  2))
-        >>> assert(    FecReceiver.validity_window(    2, 65534,  2))
-        >>> assert(not FecReceiver.validity_window(    5, 65534,  2))
-        >>> assert(    FecReceiver.validity_window(65534, 65534,  2))
-        >>> assert(    FecReceiver.validity_window(65535, 65534,  2))
+        >>> FecReceiver.validity_window(0, 5, 10)
+        False
+        >>> FecReceiver.validity_window(5, 5, 10)
+        True
+        >>> FecReceiver.validity_window(8, 5, 10)
+        True
+        >>> FecReceiver.validity_window(10, 5, 10)
+        True
+        >>> FecReceiver.validity_window(15, 5, 10)
+        False
+        >>> FecReceiver.validity_window(0, 65534, 2)
+        True
+        >>> FecReceiver.validity_window(2, 65534, 2)
+        True
+        >>> FecReceiver.validity_window(5, 65534, 2)
+        False
+        >>> FecReceiver.validity_window(65534, 65534, 2)
+        True
+        >>> FecReceiver.validity_window(65535, 65534, 2)
+        True
         """
         if end > start:
-            return current >= start and current <= end
-        else:
-            return current <= end or current >= start
+            return start <= current <= end
+        return current <= end or current >= start
